from tkinter import Toplevel

from matplotlib import pyplot
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from packetraven.tracks import LocationPacketTrack

VARIABLES = {
    'altitude': {'x': 'times', 'y': 'altitudes', 'xlabel': 'time', 'ylabel': 'altitude (m)'},
    'ascent_rate': {
        'x': 'times',
        'y': 'ascent_rates',
        'xlabel': 'time',
        'ylabel': 'ascent rate (m/s)',
    },
    'ground_speed': {
        'x': 'ground_speeds',
        'y': 'altitudes',
        'xlabel': 'ground speed (m/s)',
        'ylabel': 'altitude (m)',
    },
}


class LivePlot:
    def __init__(self, packet_tracks: {str: LocationPacketTrack}, variable: str):
        if variable not in VARIABLES:
            raise NotImplementedError(f'unsupported plotting variable "{variable}"')

        self.packet_tracks = packet_tracks
        self.variable = variable

        self.window.protocol('WM_DELETE_WINDOW', self.window.iconify)

        self.update()

    def update(self, packet_tracks: {str: LocationPacketTrack} = None):
        if packet_tracks is not None:
            self.packet_tracks.update(packet_tracks)

        if len(self.packet_tracks) > 0:
            if self.window.state() == 'iconic':
                self.window.deiconify()
            if self.window.focus_get() is None:
                self.window.focus_force()

            while len(self.axis.lines) > 0:
                self.axis.lines.pop(-1)

            for name, packet_track in self.packet_tracks.items():
                self.axis.scatter(
                    getattr(packet_track, VARIABLES[self.variable]['x']),
                    getattr(packet_track, VARIABLES[self.variable]['y']),
                    label=packet_track.name,
                    s=2,
                )

            self.axis.legend()
            pyplot.draw()

    @property
    def window(self) -> Toplevel:
        return self.figure.canvas.manager.window

    @property
    def figure(self) -> Figure:
        try:
            figure = self.__figure
            if not pyplot.fignum_exists(figure.number):
                raise RuntimeError
        except (AttributeError, RuntimeError):
            figure = self.__new_figure()
        self.__figure = figure
        pyplot.show(block=False)
        return figure

    @property
    def axis(self) -> Axes:
        try:
            axis = self.__axis
            figure = axis.figure
            if figure is not self.figure:
                raise RuntimeError
        except (AttributeError, RuntimeError):
            axis = self.__new_axis()
        self.__axis = axis
        return axis

    def __new_figure(self) -> Figure:
        x_label = VARIABLES[self.variable]['xlabel']
        y_label = VARIABLES[self.variable]['ylabel']

        return pyplot.figure(num=f'{y_label} / {x_label}')

    def __new_axis(self, figure: Figure = None) -> Axes:
        if figure is None:
            figure = self.figure

        x_label = VARIABLES[self.variable]['xlabel']
        y_label = VARIABLES[self.variable]['ylabel']

        axis = figure.add_subplot(1, 1, 1)
        axis.set_xlabel(x_label)
        axis.set_ylabel(y_label)

        return axis

    def close(self):
        pyplot.close(self.figure.number)



class DoublyLinkedList:
    """
    A linked list is a series of node objects, each with a link (object reference) to the next node in the series.
    The majority of the list is only accessible by starting at the first node ("head") and following the links forward.

    node_1 (head) -> node_2 -> node_3

    A doubly-linked list is similar to a linked list, except each node also contains a link to the previous node.
    Doubly-linked lists have an additional "tail" attribute, alongside "head".

    node_1 (head) <-> node_2 <-> node_3 (tail)
    """

    def __init__(self, sequence=None):
        """
        instantiate new doubly-linked list

        :param sequence: iterable sequence to populate list
        """

        self.head = None
        self.tail = None

        if sequence is not None:
            self.extend(sequence)

    class Node:
        """
        entry within doubly-linked list with three attributes: value, previous node, and next node
        """

        def __init__(self, value, previous_node, next_node):
            self.value = value
            self.previous_node = previous_node
            self.next_node = next_node

        def __eq__(self, other) -> bool:
            return self.value is other.value or self.value == other.value

        def __str__(self) -> str:
            return str(self.value)

        def __repr__(self) -> str:
            return f'{self.previous_node} -> [{self.value}] -> {self.next_node}'

    def append(self, value):
        """
        append given value as new tail

        :param value: value to append
        """

        new_node = self.Node(value, self.tail, None)

        if self.tail is not None:
            self.tail.next_node = new_node

        self.tail = new_node

        if self.head is None:
            self.head = self.tail

    def extend(self, sequence):
        """
        append all values in given iterable to end of list

        :param sequence: iterable sequence with which to extend
        """

        for entry in sequence:
            if type(entry) is self.Node:
                entry = entry.value

            self.append(entry)

    def insert(self, value, index: int):
        """
        insert value at given index

        :param value: value to insert
        :param index: index at which to insert value
        """

        node_at_index = self._node_at_index(index)

        if node_at_index is not None:
            new_node = self.Node(value, node_at_index.previous_node, node_at_index)

            if node_at_index.previous_node is not None:
                node_at_index.previous_node.next_node = new_node
            else:
                self.head = new_node

            node_at_index.previous_node = new_node
        else:
            if index == 0:
                self.head = self.Node(value, None, None)
                self.tail = self.head
            elif index > 0:
                self.tail = self.Node(value, self.tail, None)
            else:
                self.head = self.Node(value, None, self.head)

    def remove(self, value):
        """
        remove all instances of given value

        :param value: value to remove
        """

        current_node = self.head

        while current_node is not None:
            if current_node.value == value:
                self._remove_node(current_node)

            current_node = current_node.next_node

    def index(self, value) -> int:
        """
        first index of given value

        :param value: value to find
        :return: value index
        """

        index = 0
        current_node = self.head

        while current_node is not None:
            if current_node.value == value:
                return index

            current_node = current_node.next_node
            index += 1
        else:
            raise ValueError(f'{value} is not in list')

    def count(self, value) -> int:
        """
        number of instances of given value

        :param value: value to count
        :return: value count
        """

        return sum([1 for node_value in self if node_value == value])

    @property
    def difference(self) -> []:
        """
        differences between each value

        :return: list of differences
        """

        differences = []
        current_node = self.head

        while current_node is not None and current_node.next_node is not None:
            differences.append(current_node.next_node.value - current_node.value)
            current_node = current_node.next_node

        return differences

    def _node_at_index(self, index: int) -> Node:
        """
        Node indexing function.

        :param index: index
        :return: node at index
        """

        node_at_index = None

        if index >= 0:
            index_counter = 0
            current_node = self.head

            while current_node is not None:
                if index_counter == index:
                    node_at_index = current_node
                    break

                current_node = current_node.next_node
                index_counter += 1
        else:
            index_counter = -1
            current_node = self.tail

            while current_node is not None:
                if index_counter == index:
                    node_at_index = current_node
                    break

                current_node = current_node.previous_node
                index_counter -= 1

        if node_at_index is None:
            raise IndexError('list index out of range')

        return node_at_index

    def _remove_node(self, node: Node):
        if node.next_node is not None:
            node.next_node.previous_node = node.previous_node
        else:
            self.tail = node.previous_node

        if node.previous_node is not None:
            node.previous_node.next_node = node.next_node
        else:
            self.head = node.next_node

    def __getitem__(self, index: int):
        return self._node_at_index(index).value

    def __setitem__(self, index: int, value):
        self._node_at_index(index).value = value

    def __delitem__(self, index: int):
        self._remove_node(self._node_at_index(index))

    def __iter__(self):
        current_node = self.head

        while current_node is not None:
            yield current_node.value
            current_node = current_node.next_node

    def __reversed__(self):
        current_node = self.tail

        while current_node is not None:
            yield current_node.value
            current_node = current_node.previous_node

    def __contains__(self, item) -> bool:
        for value in self:
            if item == value:
                return True
        else:
            return False

    def __len__(self) -> int:
        length = 0
        current_node = self.head

        while current_node is not None:
            length += 1
            current_node = current_node.next_node

        return length

    def __eq__(self, other) -> bool:
        if len(self) == len(other):
            for index in range(len(self)):
                if self[index] != other[index]:
                    return False
            else:
                return True

    def __str__(self) -> str:
        return str(list(self))



from datetime import datetime, timedelta
import logging
from os import PathLike
from pathlib import Path
import re
import sys
import tkinter
from tkinter import filedialog, messagebox, simpledialog
from tkinter.ttk import Combobox, Separator
from typing import Callable

from dateutil.parser import parse
import numpy

from client import DEFAULT_INTERVAL_SECONDS
from client.retrieve import retrieve_packets
from packetraven.base import available_serial_ports, next_open_serial_port
from packetraven.connections import APRSDatabaseTable, APRSfi, APRSis, SerialTNC, TextFileTNC
from packetraven.plotting import LivePlot
from packetraven.tracks import APRSTrack
from packetraven.utilities import get_logger

LOGGER = get_logger('packetraven')


class PacketRavenGUI:
    def __init__(
            self,
            callsigns: [str] = None,
            start_date: datetime = None,
            end_date: datetime = None,
            log_filename: PathLike = None,
            output_filename: PathLike = None,
            interval_seconds: int = None,
            igate: bool = False,
            **kwargs,
    ):
        main_window = tkinter.Tk()
        main_window.title('PacketRaven')
        self.__windows = {'main': main_window}

        self.interval_seconds = (
            interval_seconds if interval_seconds is not None else DEFAULT_INTERVAL_SECONDS
        )

        self.__connection_configuration = {
            'aprs_fi': {'api_key': None},
            'tnc': {'tnc': None},
            'database': {
                'hostname': None,
                'database': None,
                'table': None,
                'username': None,
                'password': None,
            },
            'ssh_tunnel': {'ssh_hostname': None, 'ssh_username': None, 'ssh_password': None},
        }

        for section_name, section in self.__connection_configuration.items():
            section.update({key: value for key, value in kwargs.items() if key in section})

        self.igate = igate

        self.database = None
        self.aprs_is = None
        self.__connections = []

        self.__active = False
        self.__packet_tracks = {}

        self.__frames = {}
        self.__elements = {}

        self.__plots = {}

        configuration_frame = tkinter.Frame(main_window)
        configuration_frame.grid(row=main_window.grid_size()[1], column=0, pady=10)
        self.__frames['configuration'] = configuration_frame

        start_date_entry = self.__add_entry_box(
            configuration_frame, title='start_date', label='Start Date', width=22, sticky='w'
        )
        self.__add_entry_box(
            configuration_frame,
            title='end_date',
            label='End Date',
            width=22,
            row=start_date_entry.grid_info()['row'],
            column=start_date_entry.grid_info()['column'] + 1,
            sticky='e',
        )

        separator = Separator(configuration_frame, orient=tkinter.HORIZONTAL)
        separator.grid(
            row=configuration_frame.grid_size()[1],
            column=0,
            columnspan=configuration_frame.grid_size()[0] + 1,
            sticky='ew',
            pady=10,
        )

        self.__add_entry_box(
            configuration_frame,
            title='callsigns',
            label='Callsigns',
            width=55,
            columnspan=configuration_frame.grid_size()[0],
        )
        self.__file_selection_option = 'select file...'
        self.__add_combo_box(
            configuration_frame,
            title='tnc',
            label='TNC',
            options=list(available_serial_ports()) + [self.__file_selection_option],
            option_select=self.__select_tnc,
            width=52,
            columnspan=configuration_frame.grid_size()[0],
            sticky='w',
        )

        separator = Separator(configuration_frame, orient=tkinter.HORIZONTAL)
        separator.grid(
            row=configuration_frame.grid_size()[1],
            column=0,
            columnspan=configuration_frame.grid_size()[0] + 1,
            sticky='ew',
            pady=10,
        )

        self.__add_file_box(
            configuration_frame,
            title='log_file',
            file_select=self.__select_log_file,
            label='Log',
            width=52,
            columnspan=configuration_frame.grid_size()[0],
            sticky='w',
        )
        self.__add_file_box(
            configuration_frame,
            title='output_file',
            file_select=self.__select_output_file,
            label='Output',
            width=52,
            columnspan=configuration_frame.grid_size()[0],
            sticky='w',
        )

        separator = Separator(configuration_frame, orient=tkinter.HORIZONTAL)
        separator.grid(
            row=configuration_frame.grid_size()[1],
            column=0,
            columnspan=configuration_frame.grid_size()[0] + 1,
            sticky='ew',
            pady=10,
        )

        plot_label = tkinter.Label(configuration_frame, text='Plots')
        plot_label.grid(row=configuration_frame.grid_size()[1], column=0, sticky='w')

        plot_checkbox_frame = tkinter.Frame(configuration_frame)
        plot_checkbox_frame.grid(
            row=plot_label.grid_info()['row'],
            column=0,
            columnspan=configuration_frame.grid_size()[0] - 1,
        )

        plot_variables = ['altitude', 'ascent_rate', 'ground_speed']
        self.__plot_toggles = {}
        for plot_index, plot in enumerate(plot_variables):
            boolean_var = tkinter.BooleanVar()
            plot_checkbox = tkinter.Checkbutton(
                plot_checkbox_frame, text=plot, variable=boolean_var
            )
            plot_checkbox.grid(row=0, column=plot_index, padx=10)
            self.__plot_toggles[plot] = boolean_var

        separator = Separator(main_window, orient=tkinter.HORIZONTAL)
        separator.grid(row=main_window.grid_size()[1], column=0, sticky='ew')

        plot_frame = tkinter.Frame(main_window)
        plot_frame.grid(row=main_window.grid_size()[1], column=0, pady=10)
        self.__frames['plots'] = plot_frame

        plot_variables = ['altitude', 'ascent_rate', 'ground_speed']
        self.__plot_toggles = {}
        row = plot_frame.grid_size()[1]
        for plot_index, plot in enumerate(plot_variables):
            boolean_var = tkinter.BooleanVar()
            plot_checkbox = tkinter.Checkbutton(
                plot_frame, text=plot, variable=boolean_var,
            )
            plot_checkbox.grid(row=row, column=plot_index, sticky='nsew')
            self.__plot_toggles[plot] = boolean_var

        separator = Separator(main_window, orient=tkinter.HORIZONTAL)
        separator.grid(row=main_window.grid_size()[1], column=0, sticky='ew')

        control_frame = tkinter.Frame(main_window)
        control_frame.grid(row=main_window.grid_size()[1], column=0, pady=10)
        self.__frames['controls'] = control_frame

        self.__toggle_text = tkinter.StringVar()
        self.__toggle_text.set('Start')
        toggle_button = tkinter.Button(
            control_frame, textvariable=self.__toggle_text, command=self.toggle
        )
        toggle_button.grid(row=control_frame.grid_size()[1], column=0, sticky='nsew')

        self.callsigns = callsigns
        self.tncs = self.__connection_configuration['tnc']['tnc']

        if start_date is not None:
            self.start_date = start_date
        if end_date is not None:
            self.end_date = end_date

        self.log_filename = log_filename
        if self.log_filename is None:
            self.log_filename = Path('~') / 'Desktop'

        self.output_filename = output_filename
        if self.output_filename is None:
            self.output_filename = Path('~') / 'Desktop'

        self.__windows['main'].protocol('WM_DELETE_WINDOW', self.close)

        main_window.mainloop()

    @property
    def callsigns(self) -> [str]:
        callsigns = self.__elements['callsigns'].get()
        if len(callsigns) > 0:
            callsigns = [
                callsign.strip().upper()
                for callsign in re.split(',+\ *|\ +', callsigns.strip('"'))
            ]
        else:
            callsigns = None
        self.callsigns = callsigns
        return callsigns

    @callsigns.setter
    def callsigns(self, callsigns: [str]):
        if callsigns is not None:
            callsigns = ', '.join([callsign.upper() for callsign in callsigns])
        else:
            callsigns = ''
        self.replace_text(self.__elements['callsigns'], callsigns)

    @property
    def tncs(self) -> [str]:
        """ locations of TNCs parsing APRS audio into ASCII frames """
        tncs = []
        for tnc in self.__elements['tnc'].get().split(','):
            tnc = tnc.strip()
            if len(tnc) > 0:
                if tnc.upper() == 'AUTO':
                    try:
                        tnc = next_open_serial_port()
                    except OSError:
                        LOGGER.warning(f'no open serial ports')
                        tnc = None
                tncs.append(tnc)
        return tncs

    @tncs.setter
    def tncs(self, filenames: [str]):
        if filenames is None:
            filenames = []
        elif isinstance(filenames, str):
            filenames = [filenames]

        self.__connection_configuration['tnc']['tnc'] = filenames
        self.replace_text(self.__elements['tnc'], ', '.join(filenames))

    @property
    def start_date(self) -> datetime:
        start_date = self.__elements['start_date'].get()
        if len(start_date) > 0:
            start_date = parse(start_date)
        else:
            start_date = None
        return start_date

    @start_date.setter
    def start_date(self, start_date: datetime):
        if start_date is not None:
            if isinstance(start_date, str):
                start_date = parse(start_date)
            start_date = f'{start_date:%Y-%m-%d %H:%M:%S}'
        else:
            start_date = ''
        self.replace_text(self.__elements['start_date'], start_date)

    @property
    def end_date(self) -> datetime:
        end_date = self.__elements['end_date'].get()
        if len(end_date) > 0:
            end_date = parse(end_date)
        else:
            end_date = None
        return end_date

    @end_date.setter
    def end_date(self, end_date: datetime):
        if end_date is not None:
            if isinstance(end_date, str):
                end_date = parse(end_date)
            end_date = f'{end_date:%Y-%m-%d %H:%M:%S}'
        else:
            end_date = ''
        self.replace_text(self.__elements['end_date'], end_date)

    @property
    def log_filename(self) -> Path:
        filename = self.__elements['log_file'].get()
        if len(filename) > 0:
            filename = Path(filename)
            if filename.expanduser().resolve().is_dir():
                self.log_filename = filename
                filename = self.log_filename
        else:
            filename = None
        return filename

    @log_filename.setter
    def log_filename(self, filename: PathLike):
        if filename is not None:
            if not isinstance(filename, Path):
                filename = Path(filename)
            if filename.expanduser().resolve().is_dir():
                filename = filename / f'packetraven_log_{datetime.now():%Y%m%dT%H%M%S}.txt'
        else:
            filename = ''
        self.replace_text(self.__elements['log_file'], filename)

    @property
    def output_filename(self) -> Path:
        filename = self.__elements['output_file'].get()
        if len(filename) > 0:
            filename = Path(filename)
            if filename.expanduser().resolve().is_dir():
                self.output_filename = filename
                filename = self.output_filename
        else:
            filename = None
        return filename

    @output_filename.setter
    def output_filename(self, filename: PathLike):
        if filename is not None:
            if not isinstance(filename, Path):
                filename = Path(filename)
            if filename.expanduser().resolve().is_dir():
                filename = filename / f'packetraven_output_{datetime.now():%Y%m%dT%H%M%S}.txt'
        else:
            filename = ''
        self.replace_text(self.__elements['output_file'], filename)

    @property
    def active(self) -> bool:
        return self.__active

    @active.setter
    def active(self, active: bool):
        if active is not self.active:
            self.toggle()

    @property
    def packet_tracks(self) -> {str: APRSTrack}:
        return self.__packet_tracks

    def __select_log_file(self):
        self.log_filename = filedialog.asksaveasfilename(
            title='Create log file...',
            initialdir=self.log_filename.parent,
            initialfile=self.log_filename.stem,
            defaultextension='.txt',
            filetypes=[('Text', '*.txt')],
        )

    def __select_output_file(self):
        self.output_filename = filedialog.asksaveasfilename(
            title='Create output file...',
            initialdir=self.output_filename.parent,
            initialfile=self.output_filename.stem,
            defaultextension='.txt',
            filetypes=[
                ('Text', '*.txt'),
                ('GeoJSON', '*.geojson'),
                ('Keyhole Markup Language', '*.kml'),
            ],
        )

    def __select_tnc(self, event):
        if event.widget.get() == self.__file_selection_option:
            self.tncs = filedialog.askopenfilename(
                title='Select TNC text file...',
                defaultextension='.txt',
                filetypes=[('Text', '*.txt')],
            )

    def __add_combo_box(
            self,
            frame: tkinter.Frame,
            title: str,
            options: [str],
            option_select: Callable = None,
            **kwargs,
    ) -> Combobox:
        width = kwargs['width'] if 'width' in kwargs else None
        combo_box = Combobox(frame, width=width)
        combo_box['values'] = options
        if option_select is not None:
            combo_box.bind('<<ComboboxSelected>>', option_select)
        return self.__add_text_box(frame, title, text_box=combo_box, **kwargs)

    def __add_file_box(
            self, frame: tkinter.Frame, title: str, file_select: Callable, **kwargs
    ) -> tkinter.Entry:
        if 'row' not in kwargs:
            kwargs['row'] = frame.grid_size()[1]
        if 'column' not in kwargs:
            kwargs['column'] = 0
        file_box_kwargs = {
            key: value
            for key, value in kwargs.items()
            if key in ['row', 'column', 'columnspan']
        }
        if 'columnspan' in file_box_kwargs:
            file_box_kwargs['columnspan'] -= 1

        if 'label' in kwargs:
            text_label = tkinter.Label(frame, text=kwargs['label'])
            text_label.grid(row=kwargs['row'], column=kwargs['column'], sticky='w')
            file_box_kwargs['column'] += 1

        file_box_frame = tkinter.Frame(frame)
        file_box_frame.grid(**file_box_kwargs)

        file_box = tkinter.Entry(
            file_box_frame, width=kwargs['width'] if 'width' in kwargs else None
        )
        file_button = tkinter.Button(file_box_frame, text='...', command=file_select)

        file_box.pack(side='left')
        file_button.pack(side='left')

        self.__elements[title] = file_box
        return file_box

    def __add_entry_box(self, frame: tkinter.Frame, title: str, **kwargs) -> tkinter.Entry:
        if 'text_box' not in kwargs:
            kwargs['text_box'] = tkinter.Entry(
                frame, width=kwargs['width'] if 'width' in kwargs else None
            )
        return self.__add_text_box(frame, title, **kwargs)

    def __add_text_box(
            self,
            frame: tkinter.Frame,
            title: str,
            label: str,
            units: str = None,
            row: int = None,
            column: int = None,
            width: int = 10,
            text_box: tkinter.Entry = None,
            **kwargs,
    ) -> tkinter.Text:
        if row is None:
            row = frame.grid_size()[1]
        if column is None:
            column = 0

        if 'columnspan' in kwargs:
            if label is not None:
                kwargs['columnspan'] -= 1
            if units is not None:
                kwargs['columnspan'] -= 1

        if label is not None:
            text_label = tkinter.Label(frame, text=label)
            text_label.grid(row=row, column=column, sticky='w')
            column += 1

        if text_box is None:
            text_box = tkinter.Text(frame, width=width, height=1)
        text_box.grid(row=row, column=column, **kwargs)
        column += 1

        if units is not None:
            units_label = tkinter.Label(frame, text=units)
            units_label.grid(row=row, column=column)
            column += 1

        self.__elements[title] = text_box
        return text_box

    def toggle(self):
        if not self.active:
            if self.log_filename is not None:
                get_logger(LOGGER.name, self.log_filename)
            self.__elements['log_file'].configure(state=tkinter.DISABLED)

            start_date = self.start_date
            self.__elements['start_date'].configure(state=tkinter.DISABLED)

            end_date = self.end_date
            self.__elements['end_date'].configure(state=tkinter.DISABLED)

            callsigns = self.callsigns
            self.__elements['callsigns'].configure(state=tkinter.DISABLED)

            filter_message = 'retrieving packets'
            if start_date is not None and end_date is None:
                filter_message += f' sent after {start_date:%Y-%m-%d %H:%M:%S}'
            elif start_date is None and end_date is not None:
                filter_message += f' sent before {end_date:%Y-%m-%d %H:%M:%S}'
            elif start_date is not None and end_date is not None:
                filter_message += f' sent between {start_date:%Y-%m-%d %H:%M:%S} and {end_date:%Y-%m-%d %H:%M:%S}'
            if callsigns is not None:
                filter_message += f' from {len(callsigns)} callsigns: {callsigns}'
            LOGGER.info(filter_message)

            connection_errors = []
            try:
                tncs = self.tncs
                self.__elements['tnc'].configure(state=tkinter.DISABLED)
                for tnc in tncs:
                    try:
                        if 'txt' in tnc:
                            tnc = TextFileTNC(tnc, self.callsigns)
                            LOGGER.info(f'reading file {tnc.location}')
                        else:
                            tnc = SerialTNC(tnc, self.callsigns)
                            LOGGER.info(f'opened port {tnc.location}')
                        self.__connections.append(tnc)
                    except Exception as error:
                        connection_errors.append(f'TNC - {error}')
                self.tncs = [
                    connection.location
                    for connection in self.__connections
                    if isinstance(connection, SerialTNC) or isinstance(connection, TextFileTNC)
                ]

                api_key = self.__connection_configuration['aprs_fi']['api_key']
                if api_key is None:
                    api_key = simpledialog.askstring(
                        'APRS.fi API Key',
                        'enter API key for https://aprs.fi',
                        parent=self.__windows['main'],
                        show='*',
                    )
                try:
                    aprs_api = APRSfi(self.callsigns, api_key=api_key)
                    LOGGER.info(f'established connection to {aprs_api.location}')
                    self.__connections.append(aprs_api)
                    self.__connection_configuration['aprs_fi']['api_key'] = api_key
                except Exception as error:
                    connection_errors.append(f'aprs.fi - {error}')

                if (
                        'database' in self.__connection_configuration
                        and self.__connection_configuration['database']['hostname'] is not None
                ):
                    try:
                        ssh_tunnel_kwargs = {}
                        if 'ssh_tunnel' in self.__connection_configuration:
                            ssh_hostname = self.__connection_configuration['ssh_tunnel'][
                                'ssh_hostname'
                            ]
                            if ssh_hostname is not None:
                                ssh_tunnel_kwargs.update(
                                    self.__connection_configuration['ssh_tunnel']
                                )
                                if '@' in ssh_hostname:
                                    (
                                        ssh_tunnel_kwargs['ssh_username'],
                                        ssh_tunnel_kwargs['ssh_hostname'],
                                    ) = ssh_hostname.split('@', 1)
                                if (
                                        'ssh_username' not in ssh_tunnel_kwargs
                                        or ssh_tunnel_kwargs['ssh_username'] is None
                                ):
                                    ssh_username = simpledialog.askstring(
                                        'SSH Tunnel Username',
                                        f'enter username for SSH host "{ssh_tunnel_kwargs["ssh_hostname"]}"',
                                        parent=self.__windows['main'],
                                    )
                                    if ssh_username is None or len(ssh_username) == 0:
                                        raise ConnectionError('missing SSH username')
                                    ssh_tunnel_kwargs['ssh_username'] = ssh_username

                                if (
                                        'ssh_password' not in ssh_tunnel_kwargs
                                        or ssh_tunnel_kwargs['ssh_password'] is None
                                ):
                                    password = simpledialog.askstring(
                                        'SSH Tunnel Password',
                                        f'enter password for SSH user '
                                        f'"{ssh_tunnel_kwargs["ssh_username"]}"',
                                        parent=self.__windows['main'],
                                        show='*',
                                    )
                                    if password is None or len(password) == 0:
                                        raise ConnectionError('missing SSH password')
                                    ssh_tunnel_kwargs['ssh_password'] = password

                        database_kwargs = self.__connection_configuration['database']
                        if (
                                'username' not in database_kwargs
                                or database_kwargs['username'] is None
                        ):
                            database_username = simpledialog.askstring(
                                'Database Username',
                                f'enter username for database '
                                f'"{database_kwargs["hostname"]}/'
                                f'{database_kwargs["database"]}"',
                                parent=self.__windows['main'],
                            )
                            if database_username is None or len(database_username) == 0:
                                raise ConnectionError('missing database username')
                            database_kwargs['username'] = database_username

                        if (
                                'password' not in database_kwargs
                                or database_kwargs['password'] is None
                        ):
                            database_password = simpledialog.askstring(
                                'Database Password',
                                f'enter password for database user '
                                f'"{database_kwargs["username"]}"',
                                parent=self.__windows['main'],
                                show='*',
                            )
                            if database_password is None or len(database_password) == 0:
                                raise ConnectionError('missing database password')
                            database_kwargs['password'] = database_password
                        if 'table' not in database_kwargs or database_kwargs['table'] is None:
                            database_table = simpledialog.askstring(
                                'Database Table',
                                f'enter database table name',
                                parent=self.__windows['main'],
                            )
                            if database_table is None or len(database_table) == 0:
                                raise ConnectionError('missing database table name')
                            database_kwargs['table'] = database_table

                        self.database = APRSDatabaseTable(
                            **database_kwargs, **ssh_tunnel_kwargs, callsigns=self.callsigns
                        )
                        LOGGER.info(f'connected to {self.database.location}')
                        self.__connections.append(self.database)
                        self.__connection_configuration['database'].update(database_kwargs)
                        self.__connection_configuration['ssh_tunnel'].update(ssh_tunnel_kwargs)
                    except ConnectionError as error:
                        connection_errors.append(f'database - {error}')
                        self.database = None
                else:
                    self.database = None

                if self.igate:
                    try:
                        self.aprs_is = APRSis(self.callsigns)
                    except ConnectionError as error:
                        connection_errors.append(f'igate - {error}')
                        self.aprs_is = None

                if len(self.__connections) == 0:
                    connection_errors = '\n'.join(connection_errors)
                    raise ConnectionError(f'no connections started\n{connection_errors}')

                LOGGER.info(
                    f'listening for packets every {self.interval_seconds}s from {len(self.__connections)} '
                    f'connection(s): {", ".join([connection.location for connection in self.__connections])}'
                )

                for variable, enabled in self.__plot_toggles.items():
                    enabled = enabled.get()
                    if enabled and variable not in self.__plots:
                        self.__plots[variable] = LivePlot(self.__packet_tracks, variable)
                    elif not enabled and variable in self.__plots:
                        self.__plots[variable].close()
                        del self.__plots[variable]

                set_child_states(self.__frames['configuration'], tkinter.DISABLED)
                set_child_states(self.__frames['plots'], tkinter.DISABLED)

                for callsign in self.packet_tracks:
                    set_child_states(self.__windows[callsign], tkinter.DISABLED)

                self.__toggle_text.set('Stop')
                self.__active = True
            except Exception as error:
                messagebox.showerror(error.__class__.__name__, error)
                if '\n' in str(error):
                    for connection_error in str(error).split('\n'):
                        LOGGER.error(connection_error)
                else:
                    LOGGER.error(error)
                self.__active = False
                set_child_states(self.__frames['configuration'], tkinter.NORMAL)
                set_child_states(self.__frames['plots'], tkinter.NORMAL)

            self.retrieve_packets()
        else:
            for connection in self.__connections:
                connection.close()

                if type(connection) is SerialTNC:
                    LOGGER.info(f'closing port {connection.location}')

            LOGGER.info(f'closed {len(self.__connections)} connections')

            for callsign in self.packet_tracks:
                set_child_states(self.__windows[callsign], tkinter.DISABLED)
            set_child_states(self.__frames['configuration'], tkinter.NORMAL)
            set_child_states(self.__frames['plots'], tkinter.NORMAL)

            self.__toggle_text.set('Start')
            self.__active = False
            self.__connections = []

            logging.shutdown()

    def retrieve_packets(self):
        if self.active:
            try:
                current_time = datetime.now()

                existing_callsigns = list(self.packet_tracks)

                parsed_packets = retrieve_packets(
                    self.__connections,
                    self.__packet_tracks,
                    self.database,
                    self.output_filename,
                    self.start_date,
                    self.end_date,
                    logger=LOGGER,
                )

                if len(parsed_packets) > 0:
                    for variable, plot in self.__plots.items():
                        plot.update(self.packet_tracks)

                if self.aprs_is is not None:
                    self.aprs_is.send(parsed_packets)

                updated_callsigns = {packet.from_callsign for packet in parsed_packets}
                for callsign in updated_callsigns:
                    packet_track = self.packet_tracks[callsign]
                    packet_time = datetime.utcfromtimestamp(
                        (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00Z'))
                        / numpy.timedelta64(1, 's')
                    )

                    if callsign not in existing_callsigns:
                        window = tkinter.Toplevel()
                        window.title(callsign)

                        self.__add_text_box(
                            window,
                            title=f'{callsign}.source',
                            label=None,
                            width=25,
                            sticky='w',
                            columnspan=3,
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.callsign',
                            label='Callsign',
                            width=15,
                            sticky='w',
                            columnspan=2,
                        )
                        self.replace_text(self.__elements[f'{callsign}.callsign'], callsign)
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.packets',
                            label='Packet #',
                            width=15,
                            sticky='w',
                            columnspan=2,
                        )

                        self.__add_text_box(
                            window,
                            title=f'{callsign}.time',
                            label='Time',
                            width=19,
                            sticky='w',
                            row=self.__elements[f'{callsign}.source'].grid_info()['row'],
                            column=self.__elements[f'{callsign}.source'].grid_info()['column']
                                   + 4,
                            columnspan=2,
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.age',
                            label='Packet Age',
                            units='s',
                            sticky='w',
                            row=self.__elements[f'{callsign}.callsign'].grid_info()['row'],
                            column=self.__elements[f'{callsign}.callsign'].grid_info()[
                                       'column'
                                   ]
                                   + 3,
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.interval',
                            label='Interval',
                            units='s',
                            sticky='w',
                            row=self.__elements[f'{callsign}.packets'].grid_info()['row'],
                            column=self.__elements[f'{callsign}.packets'].grid_info()['column']
                                   + 3,
                        )

                        separator = Separator(window, orient=tkinter.HORIZONTAL)
                        separator.grid(
                            row=window.grid_size()[1],
                            column=0,
                            columnspan=7,
                            sticky='ew',
                            pady=10,
                        )

                        self.__add_text_box(
                            window,
                            title=f'{callsign}.coordinates',
                            label='Lat., Lon.',
                            width=15,
                            sticky='w',
                            columnspan=2,
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.distance',
                            label='Distance',
                            units='m',
                            sticky='w',
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.ground_speed',
                            label='Ground Speed',
                            units='m/s',
                            sticky='w',
                        )

                        self.__add_text_box(
                            window,
                            title=f'{callsign}.altitude',
                            label='Alt.',
                            units='m',
                            sticky='w',
                            row=self.__elements[f'{callsign}.coordinates'].grid_info()['row'],
                            column=self.__elements[f'{callsign}.coordinates'].grid_info()[
                                       'column'
                                   ]
                                   + 3,
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.ascent',
                            label='Ascent',
                            units='m',
                            sticky='w',
                            row=self.__elements[f'{callsign}.distance'].grid_info()['row'],
                            column=self.__elements[f'{callsign}.distance'].grid_info()[
                                       'column'
                                   ]
                                   + 3,
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.ascent_rate',
                            label='Ascent Rate',
                            units='m/s',
                            sticky='w',
                            row=self.__elements[f'{callsign}.ground_speed'].grid_info()['row'],
                            column=self.__elements[f'{callsign}.ground_speed'].grid_info()[
                                       'column'
                                   ]
                                   + 3,
                        )

                        separator = Separator(window, orient=tkinter.HORIZONTAL)
                        separator.grid(
                            row=window.grid_size()[1],
                            column=0,
                            columnspan=7,
                            sticky='ew',
                            pady=10,
                        )

                        self.__add_text_box(
                            window,
                            title=f'{callsign}.distance_downrange',
                            label='Downrange',
                            units='m',
                            sticky='w',
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.distance_overground',
                            label='Overground',
                            units='m',
                            sticky='w',
                        )

                        self.__add_text_box(
                            window,
                            title=f'{callsign}.maximum_altitude',
                            label='Max Alt.',
                            units='m',
                            sticky='w',
                            row=self.__elements[f'{callsign}.distance_downrange'].grid_info()[
                                'row'
                            ],
                            column=self.__elements[
                                       f'{callsign}.distance_downrange'
                                   ].grid_info()['column']
                                   + 3,
                        )
                        self.__add_text_box(
                            window,
                            title=f'{callsign}.time_to_ground',
                            label='Est. Landing',
                            units='s',
                            sticky='w',
                            row=self.__elements[f'{callsign}.distance_overground'].grid_info()[
                                'row'
                            ],
                            column=self.__elements[
                                       f'{callsign}.distance_overground'
                                   ].grid_info()['column']
                                   + 3,
                        )

                        separator = Separator(window, orient=tkinter.VERTICAL)
                        separator.grid(
                            row=0,
                            column=3,
                            rowspan=window.grid_size()[1] + 2,
                            sticky='ns',
                            padx=10,
                        )

                        window.protocol('WM_DELETE_WINDOW', window.iconify)

                        self.__windows[callsign] = window

                    window = self.__windows[callsign]

                    if window.state() == 'iconic':
                        window.deiconify()
                    if window.focus_get() is None:
                        window.focus_force()

                    set_child_states(window)

                    self.replace_text(
                        self.__elements[f'{callsign}.packets'], len(packet_track)
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.source'], packet_track[-1].source
                    )
                    self.replace_text(self.__elements[f'{callsign}.time'], f'{packet_time}')
                    self.replace_text(
                        self.__elements[f'{callsign}.altitude'],
                        f'{packet_track.coordinates[-1, 2]:.3f}',
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.coordinates'],
                        ', '.join(
                            f'{value:.3f}'
                            for value in reversed(packet_track.coordinates[-1, :2])
                        ),
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.ascent'],
                        f'{packet_track.ascents[-1]:.2f}',
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.distance'],
                        f'{packet_track.overground_distances[-1]:.2f}',
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.interval'],
                        f'{packet_track.intervals[-1]:.2f}',
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.ascent_rate'],
                        f'{packet_track.ascent_rates[-1]:.2f}',
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.ground_speed'],
                        f'{packet_track.ground_speeds[-1]:.2f}',
                    )

                    self.replace_text(
                        self.__elements[f'{callsign}.distance_downrange'],
                        f'{packet_track.distance_downrange:.2f}',
                    )
                    self.replace_text(
                        self.__elements[f'{callsign}.distance_overground'],
                        f'{packet_track.length:.2f}',
                    )

                    self.replace_text(
                        self.__elements[f'{callsign}.maximum_altitude'],
                        f'{packet_track.coordinates[:, 2].max():.2f}',
                    )

                for callsign, packet_track in self.__packet_tracks.items():
                    window = self.__windows[callsign]
                    packet_time = datetime.utcfromtimestamp(
                        (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00Z'))
                        / numpy.timedelta64(1, 's')
                    )

                    time_to_ground_box = self.__elements[f'{callsign}.time_to_ground']
                    if packet_track.time_to_ground >= timedelta(seconds=0):
                        time_to_ground_box.configure(state=tkinter.NORMAL)
                        current_time_to_ground = (
                                packet_time + packet_track.time_to_ground - current_time
                        )
                        self.replace_text(
                            time_to_ground_box,
                            f'{current_time_to_ground / timedelta(seconds=1):.2f}',
                        )
                    else:
                        self.replace_text(time_to_ground_box, '')
                        time_to_ground_box.configure(state=tkinter.DISABLED)

                    set_child_states(window, tkinter.DISABLED, [tkinter.Text])

                    packet_age_box = self.__elements[f'{callsign}.age']
                    packet_age_box.configure(state=tkinter.NORMAL)
                    self.replace_text(
                        packet_age_box,
                        f'{(current_time - packet_time) / timedelta(seconds=1):.2f}',
                    )
                    packet_age_box.configure(state=tkinter.DISABLED)

                if self.active:
                    self.__windows['main'].after(
                        int(self.interval_seconds * 1000), self.retrieve_packets
                    )
            except KeyboardInterrupt:
                self.close()
            except Exception as error:
                LOGGER.exception(f'{error.__class__.__name__} - {error}')

    @staticmethod
    def replace_text(element: tkinter.Entry, value: str):
        if isinstance(element, tkinter.Text):
            start_index = '1.0'
        else:
            start_index = 0

        if value is None:
            value = ''

        element.delete(start_index, tkinter.END)
        element.insert(start_index, value)

    def close(self):
        try:
            if self.active:
                self.toggle()
            for plot in self.__plots.values():
                plot.close()
            self.__windows['main'].destroy()
        except Exception as error:
            LOGGER.exception(f'{error.__class__.__name__} - {error}')
        sys.exit()


def set_child_states(frame: tkinter.Frame, state: str = None, types: [type] = None):
    if state is None:
        state = tkinter.NORMAL
    for child in frame.winfo_children():
        if isinstance(child, tkinter.Frame):
            set_child_states(child)
        else:
            if types is None or any(
                    isinstance(child, selected_type) for selected_type in types
            ):
                try:
                    child.configure(state=state)
                except tkinter.TclError:
                    continue



import configparser
import logging
from os import PathLike
from pathlib import Path
import sys

LOGGER_NAME_LENGTH = 17


def repository_root(path: PathLike = None) -> Path:
    if path is None:
        path = __file__
    if not isinstance(path, Path):
        path = Path(path)
    if path.is_file():
        path = path.parent
    if '.git' in (child.name for child in path.iterdir()) or path == path.parent:
        return path
    else:
        return repository_root(path.parent)


def read_configuration(filename: PathLike) -> {str: str}:
    configuration_file = configparser.ConfigParser()
    configuration_file.read(filename)
    return {
        section_name: {key: value for key, value in section.items()}
        for section_name, section in configuration_file.items()
        if section_name.upper() != 'DEFAULT'
    }


def get_logger(
        name: str,
        log_filename: PathLike = None,
        file_level: int = None,
        console_level: int = None,
        log_format: str = None,
) -> logging.Logger:
    if file_level is None:
        file_level = logging.DEBUG
    if console_level is None:
        console_level = logging.INFO
    logger = logging.getLogger(name)

    # check if logger is already configured
    if logger.level == logging.NOTSET and len(logger.handlers) == 0:
        # check if logger has a parent
        if '.' in name:
            logger.parent = get_logger(name.rsplit('.', 1)[0])
        else:
            # otherwise create a new split-console logger
            logger.setLevel(logging.DEBUG)
            if console_level != logging.NOTSET:
                if console_level <= logging.INFO:

                    class LoggingOutputFilter(logging.Filter):
                        def filter(self, rec):
                            return rec.levelno in (logging.DEBUG, logging.INFO)

                    console_output = logging.StreamHandler(sys.stdout)
                    console_output.setLevel(console_level)
                    console_output.addFilter(LoggingOutputFilter())
                    logger.addHandler(console_output)

                console_errors = logging.StreamHandler(sys.stderr)
                console_errors.setLevel(max((console_level, logging.WARNING)))
                logger.addHandler(console_errors)

    if log_filename is not None:
        if not isinstance(log_filename, Path):
            log_filename = Path(log_filename)
        log_filename = log_filename.resolve().expanduser()
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(file_level)
        for existing_file_handler in [
            handler for handler in logger.handlers if type(handler) is logging.FileHandler
        ]:
            logger.removeHandler(existing_file_handler)
        logger.addHandler(file_handler)

    if log_format is None:
        log_format = '%(asctime)s | %(levelname)-8s | %(message)s'
    log_formatter = logging.Formatter(log_format)
    for handler in logger.handlers:
        handler.setFormatter(log_formatter)

    return logger



from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
import re
from time import sleep
from typing import Any, Sequence
from urllib.parse import urlparse

import aprslib
from dateutil.parser import parse as parse_date
import requests
from serial import Serial
from shapely.geometry import Point
from tablecrow import PostGresTable
from tablecrow.table import split_URL_port

from .base import (
    APRSPacketSink,
    APRSPacketSource,
    NetworkConnection,
    PacketSink,
    PacketSource,
    next_open_serial_port,
)
from .packets import APRSPacket, LocationPacket
from .parsing import InvalidPacketError
from .utilities import get_logger, read_configuration, repository_root

LOGGER = get_logger('connection')

CREDENTIALS_FILENAME = repository_root() / 'credentials.config'


class TimeIntervalError(Exception):
    pass


class SerialTNC(APRSPacketSource):
    def __init__(self, serial_port: str = None, callsigns: [str] = None):
        """
        Connect to TNC over given serial port.

        :param serial_port: port name
        :param callsigns: list of callsigns to return from source
        """

        if serial_port is None or serial_port == '' or serial_port == 'auto':
            try:
                serial_port = next_open_serial_port()
            except ConnectionError:
                raise ConnectionError('could not find TNC connected to serial')
        else:
            serial_port = serial_port.strip('"')

        self.serial_connection = Serial(serial_port, baudrate=9600, timeout=1)
        super().__init__(self.serial_connection.port, callsigns)
        self.__last_access_time = None

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )
        packets = []
        for line in self.serial_connection.readlines():
            try:
                packet = APRSPacket.from_frame(line, source=self.location)
                packets.append(packet)
            except Exception as error:
                LOGGER.error(f'{error.__class__.__name__} - {error}')
        if self.callsigns is not None:
            packets = [packet for packet in packets if packet.from_callsign in self.callsigns]
        self.__last_access_time = datetime.now()
        return packets

    def close(self):
        self.serial_connection.close()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.location}")'


class TextFileTNC(APRSPacketSource):
    def __init__(self, filename: PathLike = None, callsigns: str = None):
        """
        read APRS packets from a given text file where each line consists of the time sent (`YYYY-MM-DDTHH:MM:SS`) followed by
        a colon `:` and then the raw APRS string

        :param filename: path to text file
        :param callsigns: list of callsigns to return from source
        """

        if not urlparse(str(filename)).scheme in ['http', 'https', 'ftp', 'sftp']:
            if not isinstance(filename, Path):
                if isinstance(filename, str):
                    filename = filename.strip('"')
                filename = Path(filename)
            filename = str(filename)

        super().__init__(filename, callsigns)
        self.__last_access_time = None
        self.__parsed_lines = []

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        if isinstance(self.location, Path):
            file_connection = open(self.location.expanduser().resolve())
            lines = file_connection.readlines()
        else:
            file_connection = requests.get(self.location, stream=True)
            lines = file_connection.iter_lines()

        packets = []
        for line in lines:
            if len(line) > 0:
                if isinstance(line, bytes):
                    line = line.decode()
                if line not in self.__parsed_lines:
                    self.__parsed_lines.append(line)
                    try:
                        packet_time, raw_aprs = line.split(': ', 1)
                        packet_time = parse_date(packet_time)
                    except:
                        raw_aprs = line
                        packet_time = datetime.now()
                    raw_aprs = raw_aprs.strip()
                    try:
                        packets.append(
                            APRSPacket.from_frame(raw_aprs, packet_time, source=self.location)
                        )
                    except Exception as error:
                        LOGGER.error(f'{error.__class__.__name__} - {error}')

        file_connection.close()

        if self.callsigns is not None:
            packets = [packet for packet in packets if packet.from_callsign in self.callsigns]
        self.__last_access_time = datetime.now()

        return packets

    def close(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.location}")'


class APRSfi(APRSPacketSource, NetworkConnection):
    interval = timedelta(seconds=10)

    def __init__(self, callsigns: [str], api_key: str = None):
        """
        connect to https://aprs.fi

        :param callsigns: list of callsigns to return from source
        :param api_key: API key for aprs.fi
        """

        url = 'https://api.aprs.fi/api/get'
        if callsigns is None:
            raise ConnectionError(f'queries to {url} require a list of callsigns')
        super().__init__(url, callsigns)

        if api_key is None or api_key == '':
            configuration = read_configuration(CREDENTIALS_FILENAME)

            if 'APRS_FI' in configuration:
                api_key = configuration['APRS_FI']['api_key']
            else:
                raise ConnectionError(f'no APRS.fi API key specified')

        if not self.connected:
            raise ConnectionError(f'no network connection')

        self.api_key = api_key

        self.__last_access_time = None

    @property
    def api_key(self) -> str:
        return self.__api_key

    @api_key.setter
    def api_key(self, api_key: str):
        response = requests.get(
            f'{self.location}?name=OH2TI&what=wx&apikey={api_key}&format=json'
        ).json()
        if response['result'] == 'fail':
            raise ConnectionError(response['description'])
        self.__api_key = api_key

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        query = {
            'name': ','.join(self.callsigns),
            'what': 'loc',
            'apikey': self.api_key,
            'format': 'json',
            'timerange': int(timedelta(days=1) / timedelta(seconds=1)),
            'tail': int(timedelta(days=1) / timedelta(seconds=1)),
        }

        query = '&'.join(f'{key}={value}' for key, value in query.items())

        response = requests.get(f'{self.location}?{query}').json()
        if response['result'] != 'fail':
            packets = []
            for packet_candidate in response['entries']:
                try:
                    packet = APRSPacket.from_frame(packet_candidate, source=self.location)
                    packets.append(packet)
                except Exception as error:
                    LOGGER.error(f'{error.__class__.__name__} - {error}')
        else:
            LOGGER.warning(f'query failure "{response["code"]}: {response["description"]}"')
            packets = []

        self.__last_access_time = datetime.now()
        return packets

    def close(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.callsigns)}, {repr(re.sub(".", "*", self.api_key))})'


class PacketDatabaseTable(PostGresTable, PacketSource, PacketSink):
    __default_fields = {
        'time': datetime,
        'x': float,
        'y': float,
        'z': float,
        'source': str,
        'point': Point,
    }

    def __init__(self, hostname: str, database: str, table: str, **kwargs):
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        if 'primary_key' not in kwargs:
            kwargs['primary_key'] = 'time'
        kwargs['fields'] = {**self.__default_fields, **kwargs['fields']}
        PostGresTable.__init__(self, hostname, database, table, **kwargs)
        PacketSource.__init__(
            self, f'postgresql://{self.hostname}:{self.port}/{self.database}/{self.name}'
        )

        if not self.connected:
            raise ConnectionError(f'cannot connect to {self.location}')

        self.__last_access_time = None
        self.__send_buffer = []

    @property
    def packets(self) -> [LocationPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )
        packets = [
            LocationPacket(**{key: value for key, value in record.items() if key != 'point'})
            for record in self.records
        ]
        self.__last_access_time = datetime.now()
        return packets

    def __getitem__(self, key: Any) -> LocationPacket:
        record = super().__getitem__(key)
        record = {key.replace('packet_', ''): value for key, value in record.items()}
        return LocationPacket(**record, crs=self.crs)

    def __setitem__(self, time: datetime, packet: LocationPacket):
        if packet.crs != self.crs:
            packet.transform_to(self.crs)
        PostGresTable.__setitem__(self, time, self.__packet_record(packet))

    def insert(self, packets: [APRSPacket]):
        for packet in packets:
            if packet.crs != self.crs:
                packet.transform_to(self.crs)
        records = [self.__packet_record(packet) for packet in packets]
        PostGresTable.insert(self, records)

    def __contains__(self, packet: LocationPacket) -> bool:
        if isinstance(packet, LocationPacket):
            packet = [packet[key.replace('packet_', '')] for key in self.primary_key]
        return PostGresTable.__contains__(self, packet)

    def send(self, packets: [LocationPacket]):
        new_packets = [packet for packet in packets if packet not in self]
        if len(self.__send_buffer) > 0:
            new_packets.extend(self.__send_buffer)
            self.__send_buffer.clear()
        if len(new_packets) > 0:
            LOGGER.info(
                f'sending {len(new_packets)} packet(s) to {self.location}: {new_packets}'
            )
            try:
                self.insert(new_packets)
            except ConnectionError as error:
                LOGGER.info(
                    f'could not send packet(s) ({error}); reattempting on next iteration'
                )
                self.__send_buffer.extend(new_packets)

    def close(self):
        self.connection.close()
        if self.tunnel is not None:
            self.tunnel.stop()

    @staticmethod
    def __packet_record(packet: LocationPacket) -> {str: Any}:
        return {
            'time': packet.time,
            'x': packet.coordinates[0],
            'y': packet.coordinates[1],
            'z': packet.coordinates[2],
            'point': Point(*packet.coordinates),
            **{f'packet_{field}': value for field, value in packet.attributes.items()},
        }


class APRSDatabaseTable(PacketDatabaseTable, APRSPacketSource, APRSPacketSink):
    __aprs_fields = {
        'from': str,
        'to': str,
        'path': [str],
        'via': str,
        'timestamp': str,
        'symbol': str,
        'symbol_table': str,
        'latitude': float,
        'longitude': float,
        'altitude': float,
        'messagecapable': str,
        'format': str,
        'gpsfixstatus': str,
        'comment': str,
        'raw': str,
    }

    def __init__(
            self, hostname: str, database: str, table: str, callsigns: [str] = None, **kwargs
    ):
        """
        Create a new database connection to a table containing APRS packet data.

        :param hostname: hostname of database
        :param database: database name
        :param table: table name
        :param callsigns: list of callsigns to return from source
        """

        if 'fields' not in kwargs:
            kwargs['fields'] = self.__aprs_fields
        if 'primary_key' not in kwargs:
            kwargs['primary_key'] = ('time', 'packet_from')
        elif 'callsign' in kwargs['primary_key']:
            kwargs['primary_key'][kwargs['primary_key'].index('callsign')] = 'packet_from'
        kwargs['fields'] = {
            f'packet_{field}': field_type for field, field_type in kwargs['fields'].items()
        }
        PacketDatabaseTable.__init__(self, hostname, database, table, **kwargs)
        location = f'postgres://{self.hostname}:{self.port}/{self.database}/{self.name}'
        APRSPacketSource.__init__(self, location, callsigns)
        APRSPacketSink.__init__(self, location)
        self.__last_access_time = None

    @property
    def packets(self) -> [APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        packets = []
        for record in self.records:
            record = {
                field
                if field in ['time', 'x', 'y', 'z']
                else field.replace('packet_', ''): value
                for field, value in record.items()
                if field not in ['point']
            }
            record['from_callsign'] = record['from']
            del record['from']
            if 'to' in record:
                record['to_callsign'] = record['to']
                del record['to']
            else:
                record['to'] = None
            if record['source'] is None:
                record['source'] = self.location
            packets.append(APRSPacket(**record))

        self.__last_access_time = datetime.now()
        return packets

    def __getitem__(self, key: (datetime, str)) -> APRSPacket:
        packet = PacketDatabaseTable.__getitem__(self, key)
        attributes = packet.attributes
        from_callsign = attributes['from']
        del attributes['from']
        if 'to' in attributes:
            to_callsign = attributes['to']
            del attributes['to']
        else:
            to_callsign = None
        return APRSPacket(
            from_callsign,
            to_callsign,
            packet.time,
            *packet.coordinates,
            packet.crs,
            **attributes,
        )

    def __setitem__(self, key: (datetime, str), packet: APRSPacket):
        PacketDatabaseTable.__setitem__(self, key, packet)

    def __contains__(self, packet: APRSPacket) -> bool:
        return PacketDatabaseTable.__contains__(self, packet)

    def send(self, packets: [APRSPacket]):
        PacketDatabaseTable.send(self, packets)

    def insert(self, packets: [APRSPacket]):
        PacketDatabaseTable.insert(self, packets)

    @property
    def records(self) -> [{str: Any}]:
        if self.callsigns is not None:
            return self.records_where({'packet_from': self.callsigns})
        else:
            return self.records_where(None)

    @staticmethod
    def __packet_record(packet: LocationPacket) -> {str: Any}:
        return {
            'time': packet.time,
            'callsign': packet.callsign,
            'x': packet.coordinates[0],
            'y': packet.coordinates[1],
            'z': packet.coordinates[2],
            'point': Point(*packet.coordinates),
            **{f'packet_{field}': value for field, value in packet.attributes.items()},
        }


class APRSis(APRSPacketSink, APRSPacketSource, NetworkConnection):
    def __init__(self, callsigns: [str] = None, hostname: str = None):
        if hostname is not None:
            self.__hostname, self.__port = split_URL_port(hostname)
        else:
            self.__hostname, self.__port = ('rotate.aprs.net', 10152)

        NetworkConnection.__init__(self, f'{self.hostname}:{self.port}')
        APRSPacketSource.__init__(self, f'{self.hostname}:{self.port}', callsigns)

        if not self.connected:
            raise ConnectionError(f'no network connection')

        self.__send_buffer = []

    @property
    def hostname(self) -> str:
        return self.__hostname

    @property
    def port(self) -> int:
        return self.__port

    def send(self, packets: [APRSPacket]):
        if not isinstance(packets, Sequence) or isinstance(packets, str):
            packets = [packets]
        packets = [
            packet if not isinstance(packet, str) else APRSPacket.from_frame(packet)
            for packet in packets
        ]

        if len(self.__send_buffer) > 0:
            packets.extend(self.__send_buffer)
            self.__send_buffer.clear()

        callsigns = {packet.from_callsign for packet in packets}
        packets = {
            callsign: [packet for packet in packets if packet.from_callsign == callsign]
            for callsign in callsigns
        }

        if len(packets) > 0:
            LOGGER.info(f'sending {len(packets)} packet(s) to {self.location}: {packets}')
            for callsign, callsign_packets in packets.items():
                try:
                    frames = [packet.frame for packet in callsign_packets]
                    aprs_is = aprslib.IS(
                        callsign, aprslib.passcode(callsign), self.hostname, self.port
                    )
                    aprs_is.connect()
                    if len(frames) > 0:
                        aprs_is.sendall(r'\rn'.join(frames))
                    aprs_is.close()
                except ConnectionError as error:
                    LOGGER.info(
                        f'could not send packet(s) ({error}); reattempting on next iteration'
                    )
                    self.__send_buffer.extend(packets)

    @property
    def packets(self) -> [APRSPacket]:
        if self.callsigns is None:
            raise ConnectionError(
                f'cannot retrieve packets from APRS-IS with no callsigns specified'
            )

        packets = []

        def add_frames(frame: str):
            try:
                packet = APRSPacket.from_frame(frame)
                if packet.from_callsign in self.callsigns and packet not in packets:
                    packets.append(packet)
            except InvalidPacketError:
                pass

        aprs_is = aprslib.IS('NOCALL', '-1', self.hostname, self.port)

        aprs_is.connect()
        aprs_is.consumer(callback=add_frames, raw=True, blocking=False)
        aprs_is.close()

        return packets

    @property
    def connected(self) -> bool:
        aprs_is = aprslib.IS('NOCALL', '-1', self.hostname, self.port)
        for _ in range(5):
            try:
                aprs_is.connect()
                aprs_is.close()
                return True
            except:
                sleep(1)
                continue
        else:
            return False

    def close(self):
        pass



#!/usr/bin/env python
from setuptools import config, find_packages, setup

try:
    from dunamai import Version
except ImportError:
    import sys
    import subprocess

    subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'dunamai'])
    from dunamai import Version

try:
    import pyproj
except ImportError:
    import platform

    if platform.system() == 'Windows':
        import subprocess

        subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'pipwin'])
        subprocess.check_call([sys.executable, '-m', 'pipwin', 'install', 'pyproj'])

metadata = config.read_configuration('setup.cfg')['metadata']

setup(
    name=metadata['name'],
    version=Version.from_any_vcs().serialize(),
    author=metadata['author'],
    author_email=metadata['author_email'],
    description=metadata['description'],
    long_description=metadata['long_description'],
    long_description_content_type='text/markdown',
    url=metadata['url'],
    packages=find_packages(),
    python_requires='>=3.8',
    setup_requires=['dunamai', 'setuptools>=41.2'],
    install_requires=[
        'aprslib',
        'haversine',
        'numpy==1.19.3',
        'pyserial',
        'geojson',
        'fastkml',
        'matplotlib',
        'pyproj',
        'psycopg2-binary',
        'requests',
        'shapely',
        'sshtunnel',
        'tablecrow',
    ],
    extras_require={'testing': ['flake8', 'pytest', 'pytest-cov'], 'development': ['oitnb']},
    entry_points={'console_scripts': ['packetraven=client.cli:main']},
)
